#include "stdlib.h"
#include "stdio.h"
#include "TEncAuf.h"
#include "math.h"
#include "assert.h"
#include "TLibCommon/TComPicYuv.h"

#if TE4_4_2_1

Int usf_luma_fixed[16][NTAPS_US_LUMA] =
{
  {  0,  0,  0, 64,  0,  0,  0,  0}, //
  { -1, -1, -1, -1, -1, -1, -1, -1}, //
  { -1, -1, -1, -1, -1, -1, -1, -1}, //
  { -1, -1, -1, -1, -1, -1, -1, -1}, // 
  { -1, -1, -1, -1, -1, -1, -1, -1}, //
  { -1, 4, -11, 52, 26,  -8, 3, -1}, // <-> actual phase shift 1/3, used for spatial scalability x1.5      
  { -1, -1, -1, -1, -1, -1, -1, -1}, //       
  { -1, -1, -1, -1, -1, -1, -1, -1}, // 
  { -1, 4, -11, 40, 40, -11, 4, -1}, // <-> actual phase shift 1/2, equal to HEVC MC, used for spatial scalability x2
  { -1, -1, -1, -1, -1, -1, -1, -1}, // 
  { -1, -1, -1, -1, -1, -1, -1, -1}, // 
  { -1, 3,  -8, 26, 52, -11, 4, -1}, // <-> actual phase shift 2/3, used for spatial scalability x1.5
  { -1, -1, -1, -1, -1, -1, -1, -1}, // 
  { -1, -1, -1, -1, -1, -1, -1, -1}, // 
  { -1, -1, -1, -1, -1, -1, -1, -1}, // 
  { -1, -1, -1, -1, -1, -1, -1, -1}  // 
};


TEncAuf::TEncAuf(void)
{
  m_pPicYuvTemp[0] = NULL;
  m_pPicYuvTemp[1] = NULL;
}

TEncAuf::~TEncAuf(void)
{
  if ( m_pPicYuvTemp[0] )
  {
    m_pPicYuvTemp[0]->destroyLuma();
    delete m_pPicYuvTemp[0];
    m_pPicYuvTemp[0] = NULL;
  }

  if ( m_pPicYuvTemp[1] )
  {
    m_pPicYuvTemp[1]->destroyLuma();
    delete m_pPicYuvTemp[1];
    m_pPicYuvTemp[1] = NULL;
  }
}

Int TEncAuf::computeAufFilters( TComPicYuv* pPicYuvOrg, TComPicYuv* pPicYuvRecBase, AUFParam* aufParam, Int iSliceInterval)
{
  Pel* pOrgY   = pPicYuvOrg->getLumaAddr();
  Int orgWidth = pPicYuvOrg->getWidth() - pPicYuvOrg->getPicCropLeftOffset() - pPicYuvOrg->getPicCropRightOffset();
  Int orgHeight = pPicYuvOrg->getHeight() - pPicYuvOrg->getPicCropTopOffset() - pPicYuvOrg->getPicCropBottomOffset();
  Int orgStride = pPicYuvOrg->getStride();
  Pel* pRefY    = pPicYuvRecBase->getLumaAddr();
  Int refWidth  = pPicYuvRecBase->getWidth() - pPicYuvRecBase->getPicCropLeftOffset() - pPicYuvRecBase->getPicCropRightOffset();
  Int refHeight = pPicYuvRecBase->getHeight() - pPicYuvRecBase->getPicCropTopOffset() - pPicYuvRecBase->getPicCropBottomOffset();
  Int refStride = pPicYuvRecBase->getStride();

  m_interWidth = refWidth;
  m_interHeight = orgHeight;
  Int dir; // hor/ver;
  Int iSpatialIdx;

  if (orgWidth ==  refWidth)
  {
    iSpatialIdx = 0;
  }
  else if (refWidth*3 ==  orgWidth*2)
  {
    iSpatialIdx = 1;
  }
  else if (refWidth*2 ==  orgWidth)
  {
    iSpatialIdx = 2;
  }
  else 
  {
    assert(0);
  }

  aufParam->setSpatialIdx (iSpatialIdx);

  if (m_pPicYuvTemp[0] == NULL)
  {
    m_pPicYuvTemp[0] = new TComPicYuv;
    m_pPicYuvTemp[0]->createLuma( orgWidth, orgHeight, g_uiMaxCUWidth, g_uiMaxCUHeight, g_uiMaxCUDepth);  
  }

  if (m_pPicYuvTemp[1] == NULL)
  {
    m_pPicYuvTemp[1] = new TComPicYuv;
    m_pPicYuvTemp[1]->createLuma( orgWidth, orgHeight, g_uiMaxCUWidth, g_uiMaxCUHeight, g_uiMaxCUDepth);  
  }

  if ( refWidth != orgWidth )
  {
    smoothOrgPic(m_pPicYuvTemp[1]->getLumaAddr(), orgWidth, orgHeight, m_pPicYuvTemp[0]->getStride(), pOrgY, orgWidth, orgHeight, orgStride, m_pPicYuvTemp[0]->getLumaAddr() );
  }

  dir = m_interHeight != orgHeight;

  if ( refWidth != orgWidth )
  {
    xDownsampleOneD(m_pPicYuvTemp[0]->getLumaAddr(), m_interWidth, m_interHeight, m_pPicYuvTemp[0]->getStride(), m_pPicYuvTemp[1]->getLumaAddr(), orgWidth, orgHeight, m_pPicYuvTemp[0]->getStride(), dir );
  }

  dir = !dir;

  if ( refWidth == orgWidth )
  {
    xGetStatistics( pOrgY, orgWidth, orgHeight, orgStride, pRefY, refWidth, refHeight, refStride, dir );
  }
  else
  {
    xGetStatistics( m_pPicYuvTemp[0]->getLumaAddr(), m_interWidth, m_interHeight, m_pPicYuvTemp[0]->getStride(), pRefY, refWidth, refHeight, refStride, dir );
  }


  xDeriveFilterCoef(dir);
  xQuantizeCoeff(dir, iSliceInterval); 

  xUpsampleOneD(m_pPicYuvTemp[0]->getLumaAddr(), m_interWidth, m_interHeight, m_pPicYuvTemp[0]->getStride(), pRefY, refWidth, refHeight, refStride, dir);

  dir = !dir;


  if ( refWidth == orgWidth )
  {
    xGetStatistics( pOrgY, orgWidth, orgHeight, orgStride,  m_pPicYuvTemp[0]->getLumaAddr(), m_interWidth, m_interHeight, m_pPicYuvTemp[0]->getStride(), dir );
  }
  else
  {
    xGetStatistics( m_pPicYuvTemp[1]->getLumaAddr(), orgWidth, orgHeight, m_pPicYuvTemp[1]->getStride(), m_pPicYuvTemp[0]->getLumaAddr(), m_interWidth, m_interHeight, m_pPicYuvTemp[0]->getStride(), dir );
  }

  xDeriveFilterCoef(dir);

  xQuantizeCoeff(dir, iSliceInterval);

  if( xSelectFilter(pPicYuvOrg, pPicYuvRecBase) )
  {
    aufParam->setAufFlag (true);
    for( Int i = 0; i < 2; i++)
    {
      for( Int j = 0; j < 16; j++)
      {
        memcpy( aufParam->getAufCoeff( i, j), &m_iCoeff[i][j][0], sizeof(Int) * NTAPS_US_LUMA );
      }
    }
  }
  else
  {
    aufParam->setAufFlag (false);
  }

  aufParam->encAufPara();

  return 1;
}

Int TEncAuf::xDeriveFilterCoef( Int dir)
{
  for ( Int filterIdx = 0 ; filterIdx < FILTER_SET; filterIdx++)
  {
    if (m_iCount[dir][filterIdx])
    {
      gnsSolveByChol(m_dAutoCorr[dir][filterIdx], m_dCrosCorr[dir][filterIdx], m_dCoeff[dir][filterIdx], NTAPS_US_LUMA);
    }
  }

  return 1;
}

void TEncAuf::xQuantizeCoeff(Int dir, Int iSliceInterval)
{
  Int filterIdx;

  for ( filterIdx = 1 ; filterIdx < 8; filterIdx++)
  {
    if (m_iCount[dir][filterIdx])
    {
      Int sum = 0;
      for (int i = 0; i < NTAPS_US_LUMA; i++ )
      {
        if(i == NTAPS_US_LUMA / 2 - 1 + (filterIdx%16+7)/16)
        {
          continue;
        }
        double sign = m_dCoeff[dir][filterIdx][i] > 0.0 ? 1.0: -1.0;
        m_dCoeff[dir][filterIdx][i] = ((int)(m_dCoeff[dir][filterIdx][i]* (1<<US_FILTER_PREC)+0.5*sign))/(double)(1<<US_FILTER_PREC);

        m_iCoeff[dir][filterIdx][i] = ((int)(m_dCoeff[dir][filterIdx][i]* (1<<US_FILTER_PREC)+0.5*sign));
        sum +=  m_iCoeff[dir][filterIdx][i];
      }
      m_iCoeff[dir][filterIdx][NTAPS_US_LUMA / 2 - 1 + (filterIdx%16+7)/16] = (1<<US_FILTER_PREC) - sum;
      m_dCoeff[dir][filterIdx][NTAPS_US_LUMA / 2 - 1 + (filterIdx%16+7)/16] = m_iCoeff[dir][filterIdx][NTAPS_US_LUMA / 2 - 1 + (filterIdx%16+7)/16] / (double)(1<<US_FILTER_PREC);
    }
    else
    {
      for (int i = 0; i < NTAPS_US_LUMA; i++ )
      {
        m_iCoeff[dir][filterIdx][i] = usf_luma_fixed[filterIdx][i];
      }
    }
  }

  for ( filterIdx = 1 ; filterIdx < 8; filterIdx++)
  {
    if (m_iCount[dir][filterIdx])
    {
      for (int i = 0; i < NTAPS_US_LUMA; i++ )
      {
        m_iCoeff[dir][16-filterIdx][i] = m_iCoeff[dir][filterIdx][NTAPS_US_LUMA-1-i] ;
        m_dCoeff[dir][16-filterIdx][i] = m_dCoeff[dir][filterIdx][NTAPS_US_LUMA-1-i];
      }
    }
  }

  filterIdx = 0;
  if (m_iCount[dir][filterIdx])
  {
    Int sum = 0;
    m_dCoeff[dir][filterIdx][NTAPS_US_LUMA-1] = m_iCoeff[dir][filterIdx][NTAPS_US_LUMA-1] = 0;
    for (int i = 0; i < NTAPS_US_LUMA/2 - 1; i++ )
    {
      double sign = (m_dCoeff[dir][filterIdx][i]+m_dCoeff[dir][filterIdx][NTAPS_US_LUMA-2-i]) > 0.0 ? 1.0: -1.0;
      m_dCoeff[dir][filterIdx][i] = ((int)((m_dCoeff[dir][filterIdx][i] + m_dCoeff[dir][filterIdx][NTAPS_US_LUMA-2-i] ) * (1<<US_FILTER_PREC)/2.0+0.5*sign))/(double)(1<<US_FILTER_PREC);
      m_dCoeff[dir][filterIdx][NTAPS_US_LUMA-2-i] = m_dCoeff[dir][filterIdx][i];

      m_iCoeff[dir][filterIdx][i] = ((int)(m_dCoeff[dir][filterIdx][i]* (1<<US_FILTER_PREC)+0.5*sign));
      m_iCoeff[dir][filterIdx][NTAPS_US_LUMA-2-i] = m_iCoeff[dir][filterIdx][i];
      sum +=  m_iCoeff[dir][filterIdx][i]*2;
    }
    m_iCoeff[dir][filterIdx][NTAPS_US_LUMA / 2 - 1] = (1<<US_FILTER_PREC) - sum;
    m_dCoeff[dir][filterIdx][NTAPS_US_LUMA / 2 - 1] = m_iCoeff[dir][filterIdx][NTAPS_US_LUMA / 2 - 1] / (double)(1<<US_FILTER_PREC);
  }
  else
  {
    for (int i = 0; i < NTAPS_US_LUMA; i++ )
    {
      m_iCoeff[dir][filterIdx][i] = usf_luma_fixed[filterIdx][i];
    }
  }

  filterIdx = 8;
  if (m_iCount[dir][filterIdx])
  {
    Int sum = 0;
    for (int i = 0; i < NTAPS_US_LUMA/2 - 1; i++ )
    {
      double sign = (m_dCoeff[dir][filterIdx][i]+m_dCoeff[dir][filterIdx][NTAPS_US_LUMA-1-i]) > 0.0 ? 1.0: -1.0;
      m_dCoeff[dir][filterIdx][i] = ((int)((m_dCoeff[dir][filterIdx][i] + m_dCoeff[dir][filterIdx][NTAPS_US_LUMA-1-i] ) * (1<<US_FILTER_PREC)/2+0.5*sign))/(double)(1<<US_FILTER_PREC);
      m_dCoeff[dir][filterIdx][NTAPS_US_LUMA-1-i] = m_dCoeff[dir][filterIdx][i];

      m_iCoeff[dir][filterIdx][i] = ((int)(m_dCoeff[dir][filterIdx][i]* (1<<US_FILTER_PREC)+0.5*sign));
      m_iCoeff[dir][filterIdx][NTAPS_US_LUMA-1-i] = m_iCoeff[dir][filterIdx][i];
      sum +=  m_iCoeff[dir][filterIdx][i]*2;
    }
    m_iCoeff[dir][filterIdx][NTAPS_US_LUMA / 2 - 1] = ((1<<US_FILTER_PREC) - sum)/2;
    m_iCoeff[dir][filterIdx][NTAPS_US_LUMA / 2] = m_iCoeff[dir][filterIdx][NTAPS_US_LUMA / 2 - 1];

    m_dCoeff[dir][filterIdx][NTAPS_US_LUMA / 2 - 1] = m_iCoeff[dir][filterIdx][NTAPS_US_LUMA / 2 - 1] / (double)(1<<US_FILTER_PREC);
    m_dCoeff[dir][filterIdx][NTAPS_US_LUMA / 2 ] = m_dCoeff[dir][filterIdx][NTAPS_US_LUMA / 2 - 1];

  }
  else
  {
    for (int i = 0; i < NTAPS_US_LUMA; i++ )
    {
      m_iCoeff[dir][filterIdx][i] = usf_luma_fixed[filterIdx][i];
    }
  }

  Int coeff_threshold = 62;
  if (iSliceInterval == 1)
    coeff_threshold += 10;

  if (m_iCoeff[dir][0][NTAPS_US_LUMA/2 -1] >= coeff_threshold)
  {
    for( Int j = 0; j < 16; j++)
    {
      memcpy( &m_iCoeff[dir][j][0], &usf_luma_fixed[j][0], sizeof(Int) * NTAPS_US_LUMA );
    }
  }
}

Bool TEncAuf::xSelectFilter( TComPicYuv* pPicYuvOrg, TComPicYuv* pPicYuvRecBase)
{
  Int sumHAD_fix = 0;
  Int sumHAD_auf = 0;

  Pel* pOrgY   = pPicYuvOrg->getLumaAddr();
  Int orgWidth = pPicYuvOrg->getWidth() - pPicYuvOrg->getPicCropLeftOffset() - pPicYuvOrg->getPicCropRightOffset();
  Int orgHeight = pPicYuvOrg->getHeight() - pPicYuvOrg->getPicCropTopOffset() - pPicYuvOrg->getPicCropBottomOffset();
  Int orgStride = pPicYuvOrg->getStride();
  Pel* pBaseY    = pPicYuvRecBase->getLumaAddr();
  Int baseWidth  = pPicYuvRecBase->getWidth() - pPicYuvRecBase->getPicCropLeftOffset() - pPicYuvRecBase->getPicCropRightOffset();
  Int baseHeight = pPicYuvRecBase->getHeight() - pPicYuvRecBase->getPicCropTopOffset() - pPicYuvRecBase->getPicCropBottomOffset();
  Int baseStride = pPicYuvRecBase->getStride();

  Pel* pDstY = m_pPicYuvTemp[0]->getLumaAddr();
  Int  dstStride = m_pPicYuvTemp[0]->getStride();
  Pel* tempY = m_pPicYuvTemp[1]->getLumaAddr();

  xUpsampleTwoD( pDstY, tempY, orgWidth, orgHeight, dstStride, pBaseY, baseWidth, baseHeight, baseStride, usf_luma_fixed, usf_luma_fixed );
  sumHAD_fix = m_pcRdCost->calcHAD         ( pOrgY, orgStride, pDstY, dstStride, orgWidth, orgHeight );

  xUpsampleTwoD( pDstY, tempY, orgWidth, orgHeight, dstStride, pBaseY, baseWidth, baseHeight, baseStride, m_iCoeff[1], m_iCoeff[0] );
  sumHAD_auf = m_pcRdCost->calcHAD         ( pOrgY, orgStride, pDstY, dstStride, orgWidth, orgHeight );

  if( sumHAD_auf >= sumHAD_fix )
  {
    return false;
  }
  else
  {
    return true;
  }
}

void TEncAuf::xUpsampleTwoD(Pel* pDstY, Pel* tempY, Int dstWidth, Int dstHeight, Int dstStride, Pel* pSrcY, Int srcWidth, Int srcHeight, Int srcStride, Int filterVer[16][8], Int filterHor[16][8] )
{
  Int   iShift = 16;
  Int   iPhaseY = 0;
  Int   iShiftM4 = iShift - 4;

  Int   iDelta     = 4 * iPhaseY; 

  Int   iAdd, iScale;

  Int interWidth = srcWidth;
  Int interStride = dstStride;
  Int interHeight = dstHeight;

  //===== vertical upsampling =====
  iAdd   = ( ( ( srcHeight * iPhaseY ) << ( iShift - 2 ) ) + ( interHeight >> 1 ) ) / interHeight+ ( 1 << ( iShift - 5 ) );

  iScale = ( ( srcHeight << iShift ) + ( interHeight >> 1 ) ) / interHeight;

  for( int i = 0; i < interWidth; i++ )
  {
    Pel* pSrc = &pSrcY[i];
    Pel* pDst = &tempY[i];
    for( int j = 0; j < interHeight; j++ )
    {
      int iRefPos16 = (( j * iScale + iAdd ) >> iShiftM4 ) - iDelta;
      int iPhase    = iRefPos16  & 15;
      int iRefPos   = iRefPos16 >>  4;

      Int temp = 0;

      for( int k = 0; k < NTAPS_US_LUMA; k++ )
      {
        int m = Clip3( 0, srcHeight - 1, iRefPos + k - 3 );
        temp += filterVer[iPhase][k] * pSrc[m*srcStride];
      }
      pDst[j*interStride] = temp;
    }
  }

  //===== horizontal upsampling =====

  iAdd   = ( ( ( interWidth * iPhaseY ) << ( iShift - 2 ) ) + ( dstWidth >> 1 ) ) / dstWidth + ( 1 << ( iShift - 5 ) );
  iScale = ( ( interWidth << iShift ) + ( dstWidth >> 1 ) ) / dstWidth;
   
  Int offset = 1<<(2*US_FILTER_PREC-1);
  Int shift = 2*US_FILTER_PREC; 
  for( int j = 0; j < dstHeight; j++ )
  {
    Pel* pSrc = &tempY[j*interStride];
    Pel* pDst = &pDstY[j*dstStride];

    for( int i = 0; i < dstWidth; i++ )
    {
      int iRefPos16 = (( i * iScale + iAdd ) >> iShiftM4 ) - iDelta;
      int iPhase    = iRefPos16  & 15;
      int iRefPos   = iRefPos16 >>  4;

      Int temp = 0;

      for( int k = 0; k < NTAPS_US_LUMA; k++ )
      {
        int m = Clip3( 0, interWidth - 1, iRefPos + k - 3 );
        temp += filterHor[iPhase][k] * pSrc[m];
      }
      pDst[i] = Clip3(0, 255, ( temp + offset ) >> shift );
    }
  }
}


Void TEncAuf::smoothOrgPic(Pel* pDstY, Int dstWidth, Int dstHeight, Int dstStride, Pel* pSrcY, Int srcWidth, Int srcHeight, Int srcStride, Pel* tempFrameBuf )
{
  assert (dstWidth == srcWidth && dstHeight == srcHeight);

  Int fitler_smooth[10] = { 2,     0,   -14,    33,    86,    33,   -14,     0,     2,     0 };

  Int *filter = &fitler_smooth[0];
  
  //vertical filter
  for( int i = 0; i < dstWidth; i++ )
  {
    Pel* pSrc = &pSrcY[i];
    Pel* pDst = &tempFrameBuf[i];
    for( int j = 0; j < dstHeight; j++ )
    {
      int temp = 0;
      for( int k = 0; k < 10; k++ )
      {
        int m = Clip3( 0, srcHeight - 1, j + k - 4);
        temp += filter[k] * pSrc[m*srcStride];
      }
      pDst[j*dstStride] = Clip3(0, 255, ( temp + 64 ) >> 7 );
    }
  }

  //===== horizontal filter =====
  for( int j = 0; j < dstHeight; j++ )
  {
    Pel* pSrc = &tempFrameBuf[j*dstStride];
    Pel* pDst = &pDstY[j*dstStride];

    for( int i = 0; i < dstWidth; i++ )
    {
      int temp = 0;
      for( int k = 0; k < 10; k++ )
      {
        int m = Clip3( 0, srcWidth - 1, i + k - 4 );
        temp += filter[k] * pSrc[m];
      }
      pDst[i] = Clip3( 0, 255, ( temp + 64 ) >> 7);
    }
  }
}

void TEncAuf::xDownsampleOneD(Pel* pDstY, Int dstWidth, Int dstHeight, Int dstStride, Pel* pSrcY, Int srcWidth, Int srcHeight, Int srcStride, Int dir)
{
  // jdong
  int filter16[2][16][12] =  // 0.9pi
  {
    {
      { 0,     5,    -6,   -10,    37,    76,    37,   -10,    -6,     5,     0,     0},
      { 0,     5,    -4,   -11,    33,    76,    40,    -9,    -7,     5,     0,     0},
      {-1,     5,    -3,   -12,    29,    75,    45,    -7,    -8,     5,     0,     0},
      {-1,     4,    -2,   -13,    25,    75,    48,    -5,    -9,     5,     1,     0},
      {-1,     4,    -1,   -13,    22,    73,    52,    -3,   -10,     4,     1,     0},
      {-1,     4,     0,   -13,    18,    72,    55,    -1,   -11,     4,     2,    -1},
      {-1,     4,     1,   -13,    14,    70,    59,     2,   -12,     3,     2,    -1},
      {-1,     3,     1,   -13,    11,    68,    62,     5,   -12,     3,     2,    -1},
      {-1,     3,     2,   -13,     8,    65,    65,     8,   -13,     2,     3,    -1},
      {-1,     2,     3,   -12,     5,    62,    68,    11,   -13,     1,     3,    -1},
      {-1,     2,     3,   -12,     2,    59,    70,    14,   -13,     1,     4,    -1},
      {-1,     2,     4,   -11,    -1,    55,    72,    18,   -13,     0,     4,    -1},
      { 0,     1,     4,   -10,    -3,    52,    73,    22,   -13,    -1,     4,    -1},
      { 0,     1,     5,    -9,    -5,    48,    75,    25,   -13,    -2,     4,    -1},
      { 0,     1,     5,    -8,    -7,    44,    75,    28,   -12,    -3,     5,     0},
      { 0,     0,     5,    -7,    -9,    40,    76,    33,   -11,    -4,     5,     0}
    },
    {
      {2,    -3,    -9,     6,    39,    58,    39,     6,    -9,    -3,     2,     0},   
      {2,    -3,    -9,     4,    38,    58,    43,     7,    -9,    -4,     1,     0},   
      {2,    -2,    -9,     2,    35,    58,    44,     9,    -8,    -4,     1,     0},   
      {1,    -2,    -9,     1,    34,    58,    46,    11,    -8,    -5,     1,     0},   
      {1,    -1,    -8,    -1,    31,    57,    47,    13,    -7,    -5,     1,     0},   
      {1,    -1,    -8,    -2,    29,    56,    49,    15,    -7,    -6,     1,     1},   
      {1,     0,    -8,    -3,    26,    55,    51,    17,    -7,    -6,     1,     1},   
      {1,     0,    -7,    -4,    24,    54,    52,    19,    -6,    -7,     1,     1},   
      {1,     0,    -7,    -5,    22,    53,    53,    22,    -5,    -7,     0,     1},   
      {1,     1,    -7,    -6,    19,    52,    54,    24,    -4,    -7,     0,     1},   
      {1,     1,    -6,    -7,    17,    51,    55,    26,    -3,    -8,     0,     1},   
      {1,     1,    -6,    -7,    15,    49,    56,    29,    -2,    -8,    -1,     1},   
      {0,     1,    -5,    -8,    13,    48,    57,    31,    -1,    -8,    -1,     1},   
      {0,     1,    -5,    -8,    11,    46,    58,    34,     1,    -9,    -2,     1},   
      {0,     1,    -4,    -8,     9,    44,    58,    35,     2,    -9,    -2,     2},   
      {0,     1,    -4,    -9,     7,    43,    58,    38,     4,    -9,    -3,     2}
    }
  };

  int iFilter;

  if (dir)
  {
    if (  1 * srcHeight >  2 * dstHeight )   
    {
      assert(0);
    }
    else if (  3 * srcHeight >  5 * dstHeight ) 
    {
      iFilter  = 1;
    }
    else if (  4 * srcHeight >  5 * dstHeight )   
    {
      iFilter  = 0;
    }
    else
    {
      assert (0);
    }
  }
  else
  {
    if (  1 * srcWidth >  2 * dstWidth )   
    {
      assert(0);
    }
    else if (  3 * srcWidth >  5 * dstWidth ) 
    {
      iFilter  = 1;
    }
    else if (  4 * srcWidth >  5 * dstWidth )   
    {
      iFilter  = 0;
    }
    else
    {
      assert (0);
    }
  }
  Int   iShift = 16;
  Int   iPhaseY = 0;
  Int   iShiftM4 = iShift - 4;

  Int   iDelta     = 4*iPhaseY;


  Int   iAdd, iScale;
  if (dir) // ver
  {
    assert(dstWidth == srcWidth);

    iAdd   = ( ( ( srcHeight * iPhaseY ) << ( iShift - 2 ) ) + ( dstHeight >> 1 ) ) / dstHeight+ ( 1 << ( iShift - 5 ) );
    iScale = ( ( srcHeight << iShift ) + ( dstHeight >> 1 ) ) / dstHeight;
  }
  else //hor
  {
    assert(dstHeight == srcHeight);

    iAdd   = ( ( ( srcWidth * iPhaseY ) << ( iShift - 2 ) ) + ( dstWidth >> 1 ) ) / dstWidth + ( 1 << ( iShift - 5 ) );
    iScale = ( ( srcWidth << iShift ) + ( dstWidth >> 1 ) ) / dstWidth;
  }

  if (dir) //===== vertical downsampling =====
  {
    for( int i = 0; i < dstWidth; i++ )
    {
      Pel* pSrc = &pSrcY[i];
      Pel* pDst = &pDstY[i];
      for( int j = 0; j < dstHeight; j++ )
      {
        int iRefPos16 = (( j * iScale + iAdd ) >> iShiftM4 ) - iDelta;
        int iPhase    = iRefPos16  & 15;
        int iRefPos   = iRefPos16 >>  4;

        int temp = 0;
        for( int k = 0; k < 12; k++ )
        {
          int m = Clip3(0, srcHeight - 1, iRefPos + k - 5 );
          temp += filter16[iFilter][iPhase][k] * pSrc[m*srcStride];
        }

        pDst[j*dstStride] = Clip3(0, 255, ( temp + 64 ) >> 7 );
      }
    }
  } 
  else//===== horizontal downsampling =====
  {
    for( int j = 0; j < dstHeight; j++ )
    {
      Pel* pSrc = &pSrcY[j*srcStride];
      Pel* pDst = &pDstY[j*dstStride];

      for( int i = 0; i < dstWidth; i++ )
      {
        int iRefPos16 = (( i * iScale + iAdd ) >> iShiftM4 ) - iDelta;
        int iPhase    = iRefPos16  & 15;
        int iRefPos   = iRefPos16 >>  4;

        int temp = 0;
        for( int k = 0; k < 12; k++ )
        {
          int m = Clip3(0, srcWidth - 1, iRefPos + k - 5 );
          temp += filter16[iFilter][iPhase][k] * pSrc[m];
        }
        pDst[i] = Clip3( 0, 255, ( temp + 64 ) >> 7 );
      }
    }
  }
}


void TEncAuf::xUpsampleOneD(Pel* pDstY, Int dstWidth, Int dstHeight, Int dstStride, Pel* pSrcY, Int srcWidth, Int srcHeight, Int srcStride, Int dir)
{
  Int   iShift = 16;
  Int   iPhaseY = 0;
  Int   iShiftM4 = iShift - 4;



  Int   iDelta     = 4 * iPhaseY; 


  Int   iAdd, iScale;
  if (dir) // ver
  {
    assert(dstWidth == srcWidth);
    iAdd   = ( ( ( srcHeight * iPhaseY ) << ( iShift - 2 ) ) + ( dstHeight >> 1 ) ) / dstHeight+ ( 1 << ( iShift - 5 ) );
    iScale = ( ( srcHeight << iShift ) + ( dstHeight >> 1 ) ) / dstHeight;
  }
  else //hor
  {
    assert(dstHeight == srcHeight);
    iAdd   = ( ( ( srcWidth * iPhaseY ) << ( iShift - 2 ) ) + ( dstWidth >> 1 ) ) / dstWidth + ( 1 << ( iShift - 5 ) );
    iScale = ( ( srcWidth << iShift ) + ( dstWidth >> 1 ) ) / dstWidth;
  }

  //  int offset = 1 << (US_FILTER_PREC - 1);
  if (dir) //===== vertical upsampling =====
  {
    for( int i = 0; i < dstWidth; i++ )
    {
      Pel* pSrc = &pSrcY[i];
      Pel* pDst = &pDstY[i];
      for( int j = 0; j < dstHeight; j++ )
      {
        int iRefPos16 = (( j * iScale + iAdd ) >> iShiftM4 ) - iDelta;
        int iPhase    = iRefPos16  & 15;
        int iRefPos   = iRefPos16 >>  4;

#if 1
        Int temp = 0;
        Int offset = 1<<(US_FILTER_PREC-1);
        for( int k = 0; k < NTAPS_US_LUMA; k++ )
        {
          int m = Clip3( 0, srcHeight - 1, iRefPos + k - 3 );
          temp += m_iCoeff[dir][iPhase][k] * pSrc[m*srcStride];
        }
        pDst[j*dstStride] = Clip3(0, 255, ( temp + offset ) >> US_FILTER_PREC );
#else
        double temp = 0.0;
        for( int k = 0; k < NTAPS_US_LUMA; k++ )
        {
          int m = Clip3( 0, srcHeight - 1, iRefPos + k - 3 );
          temp += m_dCoeff[dir][iPhase][k] * pSrc[m*srcStride];
        }
        pDst[j*dstStride] = Clip3(0, 255, (Int)( temp + 0.5) );
#endif
      }
    }
  } 
  else//===== horizontal upsampling =====
  {
    for( int j = 0; j < dstHeight; j++ )
    {
      Pel* pSrc = &pSrcY[j*srcStride];
      Pel* pDst = &pDstY[j*dstStride];

      for( int i = 0; i < dstWidth; i++ )
      {
        int iRefPos16 = (( i * iScale + iAdd ) >> iShiftM4 ) - iDelta;
        int iPhase    = iRefPos16  & 15;
        int iRefPos   = iRefPos16 >>  4;

        Int temp = 0;
        Int offset = 1<<(US_FILTER_PREC-1);
        for( int k = 0; k < NTAPS_US_LUMA; k++ )
        {
          int m = Clip3( 0, srcWidth - 1, iRefPos + k - 3 );
          temp += m_iCoeff[dir][iPhase][k] * pSrc[m];
        }
        pDst[i] = Clip3(0, 255, ( temp + offset ) >> US_FILTER_PREC );
      }
    }
  }
}

Void TEncAuf::xGetStatistics(Pel* pOrgY, Int orgWidth, Int orgHeight, Int orgStride, Pel* pRefY, Int refWidth, Int refHeight, Int refStride, Int dir)
{
  //initilize 
  Int idx;
  Int l, k; 
  Int x, y;

  Int offset = NTAPS_US_LUMA/2 - 1;

  Int   iShift = 16;
  Int   iPhase = 0;

  Int   iShiftM4 = iShift - 4;

  Int   iDelta     = 4*iPhase;

  Int   iAdd, iScale;
  if (dir) // ver
  {
    assert(refWidth == orgWidth);

    iAdd   = ( ( ( refHeight * iPhase ) << ( iShift - 2 ) ) + ( orgHeight >> 1 ) ) / orgHeight+ ( 1 << ( iShift - 5 ) );
    iScale = ( ( refHeight << iShift ) + ( orgHeight >> 1 ) ) / orgHeight;
  }
  else //hor
  {
    assert(refHeight == orgHeight);

    iAdd   = ( ( ( refWidth * iPhase ) << ( iShift - 2 ) ) + ( orgWidth >> 1 ) ) / orgWidth + ( 1 << ( iShift - 5 ) );
    iScale = ( ( refWidth << iShift ) + ( orgWidth >> 1 ) ) / orgWidth;
  }

  for(idx = 0; idx < FILTER_SET; idx++)
  {
    for(l = 0; l < NTAPS_US_LUMA; l++)
    {
      for(k = 0; k < NTAPS_US_LUMA; k++)
      {
        m_dAutoCorr[dir][idx][l][k] = 0.0;
      }
      m_dCrosCorr[dir][idx][l] = 0.0;

    }
    m_dPixAcc[dir][idx] = 0.0;
    m_iCount[dir][idx] = 0;
  }

  if (dir) // ver
  {
    for (y = NTAPS_US_LUMA; y < orgHeight-NTAPS_US_LUMA; y++)
    {
      Pel* org = pOrgY + y*orgStride;

      Int iRefPos16 = ((y*iScale + iAdd) >> iShiftM4) - iDelta;
      Int iRefPhase = iRefPos16 & 15;
      Int iRefPos   = iRefPos16 >> 4;

      Pel* ref = pRefY +iRefPos*refStride;

      for ( x = NTAPS_US_LUMA; x < orgWidth-NTAPS_US_LUMA; x++)
      {
        org ++;
        ref ++;
        for(l = 0; l < NTAPS_US_LUMA; l++)
        {
          int filter_pos_y1 = l - offset;

          for(k = 0; k < NTAPS_US_LUMA; k++)			
          {
            int filter_pos_y2 = k - offset ;

            double ac = (*(ref + filter_pos_y1*refStride ))*(*(ref+ filter_pos_y2*refStride ));

            m_dAutoCorr[dir][iRefPhase][l][k] += ac;
          }

          double cc = (*org) * (*(ref+ filter_pos_y1*refStride ));
          m_dCrosCorr[dir][iRefPhase][l] += cc;
        }
        m_dPixAcc[dir][iRefPhase] += (*org)*(*org); 
        m_iCount[dir][iRefPhase] ++;
      }
    }
  } 
  else // hor
  {
    for ( x = NTAPS_US_LUMA; x < orgWidth-NTAPS_US_LUMA; x++)
    {
      Pel* org = pOrgY + x;

      Int iRefPos16 = ((x*iScale + iAdd) >> iShiftM4) - iDelta;
      Int iRefPhase = iRefPos16 & 15;
      Int iRefPos   = iRefPos16 >> 4;

      Pel* ref = pRefY + iRefPos;
      for (y = NTAPS_US_LUMA; y < orgHeight-NTAPS_US_LUMA; y++)
      {
        org +=orgStride;
        ref +=refStride;

        for(l = 0; l < NTAPS_US_LUMA; l++)
        {
          int filter_pos_x1 = l - offset;

          for(k = 0; k < NTAPS_US_LUMA; k++)			
          {
            int filter_pos_x2 = k - offset ;

            double ac = (*(ref + filter_pos_x1 ))*(*(ref+ filter_pos_x2 ));

            m_dAutoCorr[dir][iRefPhase][l][k] += ac;
          }

          double cc = (*org) * (*(ref+ filter_pos_x1 ));
          m_dCrosCorr[dir][iRefPhase][l] += cc;
        }
        m_dPixAcc[dir][iRefPhase] += (*org)*(*org); 
        m_iCount[dir][iRefPhase] ++;
      }
    }
  }


  double backupAC[16][NTAPS_US_LUMA][NTAPS_US_LUMA];
  double backupCC[16][NTAPS_US_LUMA];
  double backupPixelACC[16];

  for (idx = 1 ; idx < 16; idx++ )
  {
    for(l = 0; l < NTAPS_US_LUMA; l++)
    {
      for(k = 0; k < NTAPS_US_LUMA; k++)			
      {
        backupAC[idx][l][k] = m_dAutoCorr[dir][idx][l][k];
      }
      backupCC[idx][l] = m_dCrosCorr[dir][idx][l];
    }
    backupPixelACC[idx] = m_dPixAcc[dir][idx];
  }

  for ( idx = 1 ; idx < 16; idx++)
  {
    if (m_iCount[dir][idx] && m_iCount[dir][16-idx] )
    {
      for(l = 0; l < NTAPS_US_LUMA; l++)
      {
        for(k = 0; k < NTAPS_US_LUMA; k++)			
        {
          m_dAutoCorr[dir][idx][l][k] = backupAC[idx][l][k] + backupAC[16-idx][NTAPS_US_LUMA-1-l][NTAPS_US_LUMA-1-k];
        }
        m_dCrosCorr[dir][idx][l] = backupCC[idx][l] + backupCC[16-idx][NTAPS_US_LUMA-1-l];
        m_dPixAcc[dir][idx] = backupPixelACC[dir] + backupPixelACC[16 -dir];
      }
    }
  }

  idx = 0;
  for(l = 0; l < NTAPS_US_LUMA; l++)
  {
    for(k = 0; k < NTAPS_US_LUMA; k++)			
    {
      backupAC[idx][l][k] = m_dAutoCorr[dir][idx][l][k];
    }
    backupCC[idx][l] = m_dCrosCorr[dir][idx][l];
  }
  backupPixelACC[idx] = m_dPixAcc[dir][idx];

  if (m_iCount[dir][idx] )
  {
    for(l = 0; l < NTAPS_US_LUMA-1; l++)
    {
      for(k = 0; k < NTAPS_US_LUMA-1; k++)			
      {
        m_dAutoCorr[dir][idx][l][k] = backupAC[idx][l][k] + backupAC[idx][NTAPS_US_LUMA-2-l][NTAPS_US_LUMA-2-k];
      }
      m_dCrosCorr[dir][idx][l] = backupCC[idx][l] + backupCC[idx][NTAPS_US_LUMA-2-l];
      m_dPixAcc[dir][idx] = backupPixelACC[dir]*2;
    }
  }


 
  m_dCrosCorr[dir][0][NTAPS_US_LUMA-1]  = 0.0;
  for(l = 0; l < NTAPS_US_LUMA; l++)
  {
    for(k = 0; k < NTAPS_US_LUMA; k++)			
    {
      if (k == NTAPS_US_LUMA - 1 || l == NTAPS_US_LUMA - 1 )
      {
        m_dAutoCorr[dir][0][l][k] = 0.0;
      }
    }
  }

}


#define ROUND(a)  (((a) < 0)? (int)((a) - 0.5) : (int)((a) + 0.5))
#define REG              0.0001
#define REG_SQR          0.0000001

//Find filter coeff related
Int TEncAuf::gnsCholeskyDec(Double inpMatr[NTAPS_US_LUMA][NTAPS_US_LUMA], Double outMatr[NTAPS_US_LUMA][NTAPS_US_LUMA], Int noEq)
{ 
  Int i, j, k;     /* Looping Variables */
  Double scale;       /* scaling factor for each row */
  Double invDiag[NTAPS_US_LUMA];  /* Vector of the inverse of diagonal entries of outMatr */

  //  Cholesky decomposition starts

  for(i = 0; i < noEq; i++)
  {
    for(j = i; j < noEq; j++)
    {
      /* Compute the scaling factor */
      scale = inpMatr[i][j];
      if ( i > 0) 
      {
        for( k = i - 1 ; k >= 0 ; k--)
        {
          scale -= outMatr[k][j] * outMatr[k][i];
        }
      }
      /* Compute i'th row of outMatr */
      if(i == j)
      {
        if(scale <= REG_SQR ) // if(scale <= 0 )  /* If inpMatr is singular */
        {
          return 0;
        }
        else
        {
          /* Normal operation */
          invDiag[i] =  1.0 / (outMatr[i][i] = sqrt(scale));
        }
      }
      else
      {
        outMatr[i][j] = scale * invDiag[i]; /* Upper triangular part          */
        outMatr[j][i] = 0.0;              /* Lower triangular part set to 0 */
      }                    
    }
  }
  return 1; /* Signal that Cholesky factorization is successfully performed */
}


Void TEncAuf::gnsTransposeBacksubstitution(Double U[NTAPS_US_LUMA][NTAPS_US_LUMA], Double rhs[], Double x[], Int order)
{
  Int i,j;              /* Looping variables */
  Double sum;              /* Holds backsubstitution from already handled rows */

  /* Backsubstitution starts */
  x[0] = rhs[0] / U[0][0];               /* First row of U'                   */
  for (i = 1; i < order; i++)
  {         /* For the rows 1..order-1           */

    for (j = 0, sum = 0.0; j < i; j++) /* Backsubst already solved unknowns */
    {
      sum += x[j] * U[j][i];
    }
    x[i] = (rhs[i] - sum) / U[i][i];       /* i'th component of solution vect.  */
  }
}

Void  TEncAuf::gnsBacksubstitution(Double R[NTAPS_US_LUMA][NTAPS_US_LUMA], Double z[NTAPS_US_LUMA], Int R_size, Double A[NTAPS_US_LUMA])
{
  Int i, j;
  Double sum;

  R_size--;

  A[R_size] = z[R_size] / R[R_size][R_size];

  for (i = R_size-1; i >= 0; i--)
  {
    for (j = i + 1, sum = 0.0; j <= R_size; j++)
    {
      sum += R[i][j] * A[j];
    }

    A[i] = (z[i] - sum) / R[i][i];
  }
}


Int TEncAuf::gnsSolveByChol(Double LHS[NTAPS_US_LUMA][NTAPS_US_LUMA], Double rhs[NTAPS_US_LUMA], Double x[NTAPS_US_LUMA], Int noEq)
{
  assert(noEq > 0);

  Double aux[NTAPS_US_LUMA];     /* Auxiliary vector */
  Double U[NTAPS_US_LUMA][NTAPS_US_LUMA];    /* Upper triangular Cholesky factor of LHS */
  Int  i, singular;          /* Looping variable */

  /* The equation to be solved is LHSx = rhs */

  /* Compute upper triangular U such that U'*U = LHS */
  if(gnsCholeskyDec(LHS, U, noEq)) /* If Cholesky decomposition has been successful */
  {
    singular = 1;
    /* Now, the equation is  U'*U*x = rhs, where U is upper triangular
    * Solve U'*aux = rhs for aux
    */
    gnsTransposeBacksubstitution(U, rhs, aux, noEq);         

    /* The equation is now U*x = aux, solve it for x (new motion coefficients) */
    gnsBacksubstitution(U, aux, noEq, x);   

  }
  else /* LHS was singular */ 
  {
    singular = 0;

    /* Regularize LHS */
    for(i=0; i < noEq; i++)
    {
      LHS[i][i] += REG;
    }
    /* Compute upper triangular U such that U'*U = regularized LHS */
    singular = gnsCholeskyDec(LHS, U, noEq);
    if ( singular == 1 )
    {
      /* Solve  U'*aux = rhs for aux */  
      gnsTransposeBacksubstitution(U, rhs, aux, noEq);   

      /* Solve U*x = aux for x */
      gnsBacksubstitution(U, aux, noEq, x);      
    }
    else
    {
      x[0] = 1.0;
      for (i = 1; i < noEq; i++ )
      {
        x[i] = 0.0;
      }
    }
  }  
  return singular;
}

#endif // TE4_4_2_1
