
#ifndef __TENCAUF__
#define __TENCAUF__

#include "TLibCommon/TComPic.h"
#include "TLibCommon/TComUpsampleFilter.h"
#include "TLibCommon/TComRdCost.h"
#include "TLibCommon/TComSlice.h"

#if TE4_4_2_1

#define FILTER_SET      16

class TEncAuf
{
private: 

  Double m_dAutoCorr[2][FILTER_SET][NTAPS_US_LUMA][NTAPS_US_LUMA]; //!< auto-correlation matrix
  Double m_dCrosCorr[2][FILTER_SET][NTAPS_US_LUMA]; //!< cross-correlation
  Double m_dPixAcc[2][FILTER_SET];
  Int m_iCount[2][FILTER_SET];

  Double m_dCoeff[2][FILTER_SET][NTAPS_US_LUMA];
  Int    m_iCoeff[2][FILTER_SET][NTAPS_US_LUMA];

  TComPicYuv* m_pPicYuvTemp[2];  // only luma is used, 0: intermediate downsampled signal, 1: smoothed original picture 

  Int    m_interWidth;           // image width of intermediate picture
  Int    m_interHeight;          // image height of intermediate picture 

  TComRdCost * m_pcRdCost;

public:
  TEncAuf(void);
  ~TEncAuf(void);

  Int computeAufFilters( TComPicYuv* pPicYuvOrg, TComPicYuv* pPicYuvRecBase, AUFParam* aufParam, Int iSliceInterval);

  Void setRDCost ( TComRdCost* pcRdCost) {m_pcRdCost = pcRdCost;}

private: 
  Int  gnsSolveByChol(Double LHS[NTAPS_US_LUMA][NTAPS_US_LUMA], Double rhs[NTAPS_US_LUMA], Double x[NTAPS_US_LUMA], int noEq);
  Void gnsBacksubstitution(Double R[NTAPS_US_LUMA][NTAPS_US_LUMA], Double z[NTAPS_US_LUMA], int R_size, Double A[NTAPS_US_LUMA]);
  Void gnsTransposeBacksubstitution(Double U[NTAPS_US_LUMA][NTAPS_US_LUMA], Double rhs[], Double x[],int order);
  Int  gnsCholeskyDec(Double inpMatr[NTAPS_US_LUMA][NTAPS_US_LUMA], Double outMatr[NTAPS_US_LUMA][NTAPS_US_LUMA], int noEq);

  Int  xDeriveFilterCoef ( Int dir );
  Void xGetStatistics( Pel* pOrgY, Int orgWidth, Int orgHeight, Int orgStride, Pel* pRefY, Int refWidth, Int refHeight, Int refStride, Int dir);

  Void xUpsampleOneD(Pel* pDstY, Int dstWidth, Int dstHeight, Int dstStride, Pel* pSrcY, Int srcWidth, Int srcHeight, Int srcStride, Int dir);
  Void xDownsampleOneD(Pel* pDstY, Int dstWidth, Int dstHeight, Int dstStride, Pel* pSrcY, Int srcWidth, Int srcHeight, Int srcStride, Int dir);

  Void xQuantizeCoeff(Int dir, Int iSliceInterval);

  Bool xSelectFilter    ( TComPicYuv* pPicYuvOrg, TComPicYuv* pPicYuvRecBase );
  Void xUpsampleTwoD    ( Pel* pDstY, Pel*tempY, Int orgWidth, Int orgHeight, Int orgStride, Pel* pSrcY, Int srcWidth, Int srcHeight, Int srcStride, Int filterV[16][8], Int filterH[16][8] );
 
  Void smoothOrgPic(Pel* pDstY, Int dstWidth, Int dstHeight, Int dstStride, Pel* pSrcY, Int srcWidth, Int srcHeight, Int srcStride, Pel* tempFrameBuf );

};

#endif // TE4_4_2_1
#endif //__TENCAUF__

